using System;
using System.Collections.Generic;
using System.Text;
using System.Diagnostics;
using InfiniTec.Threading;
using System.ComponentModel;
using System.Runtime.CompilerServices;
using System.Reflection;
using System.Net.Mail;
using System.Text.RegularExpressions;

namespace InfiniTec.Exchange.WebDav
{
    /// <summary>
    /// This class provides access to properties common to all Exchange items.
    /// </summary>
    [SchemaClass(typeof(NamedProperties.Item))]
    public abstract class TypedItemBase
    {
        private static string[] _ItemFields;

        private static Regex _EmailAddressParser = new Regex(@"(\""
                (?<name>.*?)\"")?  # Parse the mail address, encapsulated in quotation marks
                \s*?
                (\<(?<address>.*?)\>) # Parse the email address",
         RegexOptions.Compiled | RegexOptions.CultureInvariant | RegexOptions.ExplicitCapture | 
            RegexOptions.IgnorePatternWhitespace);

        /// <summary>
        /// This event is called when an async refresh operation has completed.
        /// </summary>
        public event AsyncCompletedEventHandler RefreshCompleted
        {
            add { BaseItem.RefreshCompleted += value; }
            remove { BaseItem.RefreshCompleted -= value; }
        }


        /// <summary>
        /// This event is called when an async save operation has completed.
        /// </summary>
        public event AsyncCompletedEventHandler SaveCompleted
        {
            add { BaseItem.SaveCompleted += value; }
            remove { BaseItem.SaveCompleted -= value; }
        }

        /// <summary>
        /// This event is called when an async delete operation has completed.
        /// </summary>
        public event AsyncCompletedEventHandler DeleteCompleted
        {
            add { BaseItem.DeleteCompleted += value; }
            remove { BaseItem.DeleteCompleted -= value; }
        }

        private WebDav.Item _BaseItem;
        private List<string> _RequiredFields;
        private Body _Body;


        /// <summary>
        /// Gets or sets the base <see cref="WebDav.Item"/>.
        /// </summary>
        /// <value>The base item.</value>
        protected WebDav.Item BaseItem
        {
            [DebuggerStepThrough]
            get { return _BaseItem; }
            [DebuggerStepThrough]
            set { _BaseItem = value; }
        }

        /// <summary>
        /// Sets the content class of the item.
        /// </summary>
        /// <param name="outlookClass">The Outlook class for the item (e.g. IPM.Note)</param>
        /// <param name="contentClass">The WebDAV class for item (e.g. urn:content-classes:message)</param>
        public void SetContentClass(string outlookClass, string contentClass)
        {
            ItemClass = outlookClass;
            ContentClass = contentClass;
        }

        /// <summary>
        /// Proovides access to all properties of the item.
        /// </summary>
        /// <value>The properties.</value>
        public PropertyCollection Properties
        {
            get { return BaseItem.Properties; }
        }

        /// <summary>
        /// Gets the <see cref="WebDav.AttachmentCollection"/> for this item.
        /// </summary>
        /// <value>The attachments.</value>
        public WebDav.AttachmentCollection Attachments
        {
            get { return BaseItem.Attachments; }
        }


        /// <summary>
        /// This method is called before the <see cref="Refresh"/> or <see cref="RefreshAsync"/>
        /// operation is called.
        /// </summary>
        protected virtual void PreRefresh()
        {
            _Body = null;

            foreach (string property in _RequiredFields)
            {
                if (!BaseItem.PropertiesToLoad.Contains(property))
                    BaseItem.PropertiesToLoad.Add(property);
            }
        }
        /// <summary>
        /// This method is called before the <see cref="Save"/> or <see cref="SaveAsync"/>
        /// operation is called.
        /// </summary>
        protected virtual void PreSave()
        {
        }

        /// <summary>
        /// Synchronously refreshes the properties of the curren item.
        /// </summary>
        public void Refresh()
        {
            PreRefresh();
            BaseItem.Refresh();
        }
        /// <summary>
        /// Asynchronously refreshes the properties of the curren item.
        /// </summary>
        public void RefreshAsync()
        {
            PreRefresh();
            BaseItem.RefreshAsync();
        }
        /// <summary>
        /// Synchronously saves the properties of the curren item.
        /// </summary>
        public void Save()
        {
            PreSave();
            BaseItem.Save();
        }
        /// <summary>
        /// Asynchronously saves the properties of the curren item.
        /// </summary>
        public void SaveAsync()
        {
            PreSave();
            BaseItem.SaveAsync();
        }
        /// <summary>
        /// Synchronously deletes the properties of the curren item.
        /// </summary>
        public void Delete()
        {
            BaseItem.Delete();
        }
        /// <summary>
        /// Asynchronously deletes the properties of the curren item.
        /// </summary>
        public void DeleteAsync()
        {
            BaseItem.DeleteAsync();
        }

        /// <summary>
        /// Initializes this class
        /// </summary>
        static TypedItemBase()
        {
            _ItemFields = ExtractFieldValues(typeof(NamedProperties.Item));
        }


        /// <summary>
        /// Initializes a new instance of the <see cref="ItemBase"/> class.
        /// </summary>
        /// <param name="path">The path.</param>
        /// <param name="connection">The connection.</param>
        public TypedItemBase(string path, Connection connection)
        {
            BaseItem = new WebDav.Item(path, connection);
            _RequiredFields = new List<string>();

            AddRequiredFields(_ItemFields);
        }

        /// <summary>
        /// Adds the required fields which to the list of fields which are
        /// loaded even if no other fields are specified.
        /// </summary>
        /// <param name="fields">The fields.</param>
        protected void AddRequiredFields(IEnumerable<string> fields)
        {
            _RequiredFields.AddRange(fields);
        }

        internal static string[] ExtractFieldValues(Type t)
        {
            List<string> result;

            result = new List<string>();
            foreach (FieldInfo info in t.GetFields(BindingFlags.Public | BindingFlags.Static))
            {
                result.Add((string)info.GetValue(null));
            }

            return result.ToArray();
        }

        /// <summary>
        /// Gets the property value.
        /// </summary>
        /// <param name="name">The name of the property.</param>
        /// <typeparam name="T">The type of the requested property.</typeparam>
        /// <returns>The value of the requested property. If not found, the default
        /// value for the specified type is returned.</returns>
        protected internal T GetPropertyValue<T>(string name)
        {
            return GetPropertyValue<T>(name, default(T));
        }
        /// <summary>
        /// Gets the property value.
        /// </summary>
        /// <param name="name">The name.</param>
        /// <typeparam name="T">The type of the requested property.</typeparam>
        /// <returns>The value of the requested property. If not found, the default
        /// value for the specified type is returned.</returns>
        protected internal T? GetNullablePropertyValue<T>(string name)
            where T: struct
        {
            Property<T> prop;

            if (!BaseItem.Properties.TryGetProperty<T>(name, out prop)) return null;
            else return prop.Value;
        }
        /// <summary>
        /// Gets the multi-valued property value.
        /// </summary>
        /// <param name="name">The name.</param>
        /// <typeparam name="T">The type of the requested property.</typeparam>
        /// <returns>The value of the requested property. If not found, the default
        /// value for the specified type is returned.</returns>
        protected internal IList<T> GetMultivaluedPropertyValue<T>(string name)
        {
            return GetMultivaluedPropertyValue<T>(name, null);
        }
        /// <summary>
        /// Gets the value of the property.
        /// </summary>
        /// <param name="name">The name of the property.</param>
        /// <param name="defaultValue">The default value for the property</param>
        /// <typeparam name="T">The type of the requested property.</typeparam>
        /// <returns>The value of the requested property. If not found, specified
        /// default value is returned.</returns>
        protected internal T GetPropertyValue<T>(string name, T defaultValue)
        {
            Property<T> prop;

            if (!BaseItem.Properties.TryGetProperty<T>(name, out prop)) return defaultValue;
            else return prop.Value;
        }
        /// <summary>
        /// Gets the property value.
        /// </summary>
        /// <param name="name">The name of the property.</param>
        /// <param name="defaultValue">The default value for the property</param>
        /// <typeparam name="T">The type of the requested property.</typeparam>
        /// <returns>The value of the requested property. If not found, the specified default
        /// value is returned.</returns>
        protected internal IList<T> GetMultivaluedPropertyValue<T>(string name, T[] defaultValue)
        {
            MultiValuedProperty<T> prop;

            if (!BaseItem.Properties.TryGetMultiValuedProperty<T>(name, out prop)) return defaultValue;
            else return prop.Values;
        }
        /// <summary>
        /// Gets the property value.
        /// </summary>
        /// <param name="name">The name.</param>
        /// <param name="value">The vlaue of the value.</param>
        /// <typeparam name="T">The type of the requested property.</typeparam>
        /// <returns>The value of the requested property. If not found, the default
        /// value for the specified type is returned.</returns>
        protected internal void SetPropertyValue<T>(string name, T value)
        {
            Property<T> prop;

            if (value == null) BaseItem.Properties.Remove(name);

            if (!BaseItem.Properties.TryGetProperty<T>(name, out prop))
            {
                prop = new Property<T>(name, value);
                BaseItem.Properties.Add(prop);
            }
            else
            {
                prop.Value = value;
            }
        }

        /// <summary>
        /// Sets the nullable property value.
        /// </summary>
        /// <typeparam name="T">The type of the property</typeparam>
        /// <param name="name">The name of the property.</param>
        /// <param name="value">The value of the property.</param>
        protected internal void SetNullablePropertyValue<T>(string name, T? value)
            where T: struct
        {
            Property<T> prop;

            if (value == null) BaseItem.Properties.Remove(name);

            if (!BaseItem.Properties.TryGetProperty<T>(name, out prop))
            {
                prop = new Property<T>(name, value.Value);
                BaseItem.Properties.Add(prop);
            }
            else
            {
                prop.Value = value.Value;
            }
        }


        /// <summary>
        /// Ensures that the property with the specified name and type is present in the 
        /// and can be accessed.
        /// </summary>
        /// <param name="name">The name of the property.</param>
        protected void EnsureMultiValuedProperty<T>(string name)
        {
            if (Properties.Contains(name)) return;

            Properties.AddMultiValued<T>(name, (T[])null);
        }


        [MethodImpl(MethodImplOptions.Synchronized)]
        private void EnsureBodyIsLoaded()
        {
            if (_Body == null) _Body = new Body(this);
        }

        /// <summary>
        /// Returns wheter the specified flags are set in the given value.
        /// </summary>
        /// <param name="value">The value to check.</param>
        /// <param name="flags">The flags to check</param>
        /// <returns>Returns true, if the flags are set in the specified value. False otherwise. If 
        /// the specified value is null, null is returned.</returns>
        protected internal bool? CheckFlagValue(int? value, int flags)
        {
            if (value == null) return null;

            return ((value.Value & flags) == flags);
        }

        [MethodImpl(MethodImplOptions.Synchronized)]
        private void EnsureMultivaluedPropertyExists<T>(string name)
        {
            if (BaseItem.Properties.Contains(name)) return;

            BaseItem.Properties.Add(new MultiValuedProperty<T>(name));
        }

        internal void LoadAddress(out EmailAddress address, string propertyName)
        {
            List<EmailAddress> addresses;

            addresses = ParseAddresses(GetPropertyValue<string>(propertyName));

            address = (addresses.Count > 0) ? addresses[0] : null;
        }

        private List<EmailAddress> ParseAddresses(string addresses)
        {
            List<EmailAddress> result;
            string name;
            string address;

            result = new List<EmailAddress>();

            if (string.IsNullOrEmpty(addresses)) return result;

            foreach (Match match in _EmailAddressParser.Matches(addresses))
            {
                name = (match.Groups["name"].Success) ? match.Groups["name"].Value : "";
                address = (match.Groups["address"].Success) ? match.Groups["address"].Value : "";

                if (address == null) continue;

                result.Add(new EmailAddress(name, address));
            }
            return result;
        }

        internal void LoadAddressList(out ChangeNotifyCollection<EmailAddress> recipients, string propertyName)
        {
            recipients = new ChangeNotifyCollection<EmailAddress>();

            recipients.AddRange(ParseAddresses(GetPropertyValue<string>(propertyName)));
        }

        internal void CommitAddressList(ChangeNotifyCollection<EmailAddress> list, string property)
        {
            StringBuilder sb;

            if (!list.HasChanged) return;

            sb = new StringBuilder();
            foreach (EmailAddress address in list)
            {
                sb.Append(address.ToString());
                sb.Append(", ");
            }

            SetPropertyValue<string>(property, sb.ToString());
        }

        /// <summary>
        /// Gets or sets the subject.
        /// </summary>
        /// <value>The subject.</value>
        public string Subject
        {
            get { return GetPropertyValue<string>(NamedProperties.Item.Subject); }
            set { SetPropertyValue<string>(NamedProperties.Item.Subject, value); }
        }

        /// <summary>
        /// Gets the item class of the item (e.g. IPM.Note).
        /// </summary>
        /// <value>The subject.</value>
        public string ItemClass
        {
            get { return GetPropertyValue<string>(NamedProperties.Item.ItemClass); }
            private set { SetPropertyValue<string>(NamedProperties.Item.ItemClass, value); }
        }
        
        /// <summary>
        /// Gets the content class of the item (e.g. IPM.Note).
        /// </summary>
        /// <value>The subject.</value>
        public string ContentClass
        {
            get { return GetPropertyValue<string>(NamedProperties.Item.ContentClass); }
            private set { SetPropertyValue<string>(NamedProperties.Item.ContentClass, value); }
        }

        /// <summary>
        /// Gets or sets the sensitivity.
        /// </summary>
        /// <value>The sensitivity.</value>
        public Sensitivity? Sensitivity
        {
            get { return (Sensitivity?) GetNullablePropertyValue<int>(NamedProperties.Item.Sensitivity); }
            set { SetNullablePropertyValue<int>(NamedProperties.Item.Sensitivity, (int) value); }
        }

        /// <summary>
        /// Gets the body.
        /// </summary>
        /// <value>The body.</value>
        public Body Body
        {
            get
            {
                EnsureBodyIsLoaded();
                return _Body;
            }
        }
        /// <summary>
        /// Gets the received date and time.
        /// </summary>
        /// <value>The received.</value>
        public DateTime? Received
        {
            get { return GetNullablePropertyValue<DateTime>(NamedProperties.Item.Received); }
        }

        /// <summary>
        /// Gets the size.
        /// </summary>
        /// <value>The size.</value>
        public int? Size
        {
            get { return GetNullablePropertyValue<int>(NamedProperties.Item.MessageSize); }
        }

        /// <summary>
        /// Gets the categories.
        /// </summary>
        /// <value>The categories.</value>
        public IList<string> Categories
        {
            get
            {
                EnsureMultivaluedPropertyExists<string>(NamedProperties.Item.Categories);

                return GetMultivaluedPropertyValue<string>(NamedProperties.Item.Categories);
            }
        }

        /// <summary>
        /// Gets or sets the importance.
        /// </summary>
        /// <value>The importance.</value>
        public Importance? Importance
        {
            get { return (Importance?) GetNullablePropertyValue<int>(NamedProperties.Item.Importance); }
            set { SetNullablePropertyValue<int>(NamedProperties.Item.Importance, (int) value); }
        }

        /// <summary>
        /// Gets or sets the in reply to.
        /// </summary>
        /// <value>The in reply to.</value>
        public string InReplyTo
        {
            get { return GetPropertyValue<string>(NamedProperties.Item.InReplyTo); }
            set { SetPropertyValue<string>(NamedProperties.Item.InReplyTo, value); }
        }


        /// <summary>
        /// Gets a value indicating whether this instance has been submitted.
        /// </summary>
        /// <value>
        /// 	<see langword="true"/> if this instance has been submitted; otherwise, <see langword="false"/>.
        /// </value>
        public bool? IsSubmitted
        {
            get { return CheckFlagValue(GetNullablePropertyValue<int>(NamedProperties.Item.MessageFlags), (int)MessageFlags.HasBeenSubmitted); }
        }


        /// <summary>
        /// Gets a value indicating whether this instance is a draft.
        /// </summary>
        /// <value>
        /// 	<see langword="true"/> if this instance is a draft; otherwise, <see langword="false"/>.
        /// </value>
        public bool? IsDraft
        {
            get { return CheckFlagValue(GetNullablePropertyValue<int>(NamedProperties.Item.MessageFlags), (int)MessageFlags.IsDraft); }
        }


        /// <summary>
        /// Returns true, if the item has been resent.
        /// </summary>
        /// <value>True, if the item has been resent. False, otherwise</value>
        public bool? IsResend
        {
            get { return CheckFlagValue(GetNullablePropertyValue<int>(NamedProperties.Item.MessageFlags), (int)MessageFlags.IsResend); }
        }

        /// <summary>
        /// Gets a value indicating whether this instance is unmodified.
        /// </summary>
        /// <value>
        /// 	<see langword="true"/> if this instance is unmodified; otherwise, <see langword="false"/>.
        /// </value>
        public bool? IsUnmodified
        {
            get { return CheckFlagValue(GetNullablePropertyValue<int>(NamedProperties.Item.MessageFlags), (int)MessageFlags.IsUnmodified); }
        }

        /// <summary>
        /// Gets the message headers.
        /// </summary>
        /// <value>The message headers.</value>
        public IDictionary<string, string> MessageHeaders
        {
            get { throw new NotImplementedException(); }
        }

        /// <summary>
        /// Gets the sent date and time.
        /// </summary>
        /// <value>The sent date and time.</value>
        public DateTime? Sent
        {
            get { return GetNullablePropertyValue<DateTime>(NamedProperties.Item.SentDate); }
        }

        /// <summary>
        /// Gets the created date and time.
        /// </summary>
        /// <value>The created date and time.</value>
        public DateTime? Created
        {
            get { return GetNullablePropertyValue<DateTime>(NamedProperties.Item.CreationDate); }
        }

        /// <summary>
        /// Gets or sets the reminder due by.
        /// </summary>
        /// <value>The reminder due by.</value>
        public DateTime? ReminderDueBy
        {
            get { return GetNullablePropertyValue<DateTime>(NamedProperties.Item.ReminderDueBy); }
            set { SetNullablePropertyValue<DateTime>(NamedProperties.Item.ReminderDueBy, value); }
        }


        /// <summary>
        /// Gets a value indicating whether a reminder is set.
        /// </summary>
        /// <value>
        /// 	<see langword="true"/> if a reminder is set; otherwise, <see langword="false"/>.
        /// </value>
        public bool? ReminderIsSet {
            get { return GetNullablePropertyValue<bool>(NamedProperties.Item.ReminderIsSet); }
        }

        /// <summary>
        /// Gets or sets the reminder minutes before start.
        /// </summary>
        /// <value>The reminder minutes before start.</value>
        public int? ReminderMinutesBeforeStart
        {
            get { return GetNullablePropertyValue<int>(NamedProperties.Item.ReminderMinutesBeforeState); }
            set { SetNullablePropertyValue<int>(NamedProperties.Item.ReminderMinutesBeforeState, value); }
        }


        /// <summary>
        /// Gets a value indicating whether this instance has attachments.
        /// </summary>
        /// <value>
        /// 	<see langword="true"/> if this instance has attachments; otherwise, <see langword="false"/>.
        /// </value>
        public bool? HasAttachments
        {
            get { return GetNullablePropertyValue<bool>(NamedProperties.Item.HasAttachments); }
        }
        /// <summary>
        /// Gets the culture of the item.
        /// </summary>
        /// <value>The culture of the item.</value>
        public string Culture
        {
            get { throw new NotImplementedException(); }
        }

        /// <summary>
        /// Gets or sets the last verb executed.
        /// </summary>
        /// <value>The last verb executed.</value>
        public LastVerbExcecuted? LastVerbExecuted
        {
            get { return (LastVerbExcecuted?) GetNullablePropertyValue<int>(NamedProperties.Item.LastVerbExecuted); }
            set { SetNullablePropertyValue<int>(NamedProperties.Item.LastVerbExecuted, (int?)value); }
        }

        /// <summary>
        /// Gets or sets the last verb execution time.
        /// </summary>
        /// <value>The last verb execution time.</value>
        public DateTime? LastVerbExecutionTime
        {
            get { return GetNullablePropertyValue<DateTime>(NamedProperties.Item.LastVerbExecutedOn); }
            set { SetNullablePropertyValue<DateTime>(NamedProperties.Item.LastVerbExecutedOn, value); }
        }

        /// <summary>
        /// Gets or sets the index of the icon.
        /// </summary>
        /// <value>The index of the icon.</value>
        public IconIndex? IconIndex
        {
            get { return (IconIndex?) GetNullablePropertyValue<int>(NamedProperties.Item.IconIndex); }
            set { SetNullablePropertyValue<int>(NamedProperties.Item.IconIndex, (int?) value); }
        }
    }
}
